import { PartialTweet } from "./TwitterTypes";
export declare const TweetSearcher: {
    validators: TweetSearchValidator[];
    static_validators: {
        [staticName: string]: TweetSearchStaticValidator;
    };
    /**
     * Search into {tweets} parameters using {query}.
     * Query is converted to regex if {is_regex}==true,
     * after keywords are removed from the query.
     *
     * Default available keywords are:
     * - `since:{YYYY-MM-DD}`
     * - `until:{YYYY-MM-DD}`
     * - `from:{screen_name|screen_name_2|screen_name_3}`
     * - `retweet_of:{screen_name} (use of | is allowed, like from:)`
     *
     * To be in the result array, a tweet must validate ALL the keywords.
     *
     * You can add your custom validators by pushing
     * to **this.validators** a **TweetSearchValidator** variable.
     *
     * @param tweets Partial tweets array.
     * @param query String to be searched for in the tweet.text
     * @param is_regex If the string should be considered as a regex during text.match or not.
     * @param static_validators Array of static validators names that should used.
     *
     * @throws Format "{keyword}: Invalid query" if user input is invalid
     * for a specific validator.
     *
     * @throws "Validator {name} does not exists" when a invalid static validator is used.
     */
    searchIntoTweets(tweets: PartialTweet[], query: string, is_regex?: boolean, static_validators?: string[]): PartialTweet[];
    reset(): void;
};
export interface TweetSearchValidator {
    /**
     * Keyword (word used to identify you keyword, before the **:**.
     * Will be interpoled into search regex.
     * Should **NOT** contain parenthesis !
     */
    keyword: string;
    /**
     * Method that generate a validator function with user input in your **query**.
     * This method **should** return a *function*.
     *
     * You should verify user-input stored inside {user_query}.
     *
     * If the return type is null-ish (`undefined`, `null`, `false`,...), search will be **aborted**
     * with **unwell-formed query error**.
     */
    validator: (user_query: string) => ((tweet: PartialTweet) => boolean);
}
export declare type TweetSearchStaticValidator = (tweet: PartialTweet) => boolean;
